@@ -7,6 +7,7 @@ import time
 import json
 import random
 import base64
+import signal
 import datetime
 import urllib.request
 
@@ -35,7 +36,7 @@ from werkzeug.security import generate_password_hash
 from . import util
 
 from .db import db
-from .models import User, Trigger, RegToken, Banlist, Setting, Authenticator, OTPToken
+from .models import User, Trigger, RegToken, Banlist, Setting, Authenticator, OTPToken, TriggerUserMap
 
 from .fido2.webauthn import PublicKeyCredentialRpEntity
 from .fido2.client import ClientData
@@ -50,7 +51,9 @@ app.config.from_object('hass_wh_triggers.config.Config')
 try:
     app.config.from_envvar('APP_CONFIG_FILE')
 except:
-    print("Using default configuration")
+    app.logger.info("Using default configuration")
+if app.config.get('PREFIX'):
+    app.wsgi_app = util.ReverseProxied(app.wsgi_app, script_name=app.config['PREFIX'])
 db.init_app(app)
 with app.app_context():
     db.create_all()
@@ -80,6 +83,7 @@ ORIGIN = app.config['ORIGIN']
 RP = PublicKeyCredentialRpEntity(RP_ID, "HASS-WH-Triggers")
 server = Fido2Server(RP)
 
+VERSION = "0.0.6"
 if isinstance(app.secret_key, bytes):
     ENCRYPTION_KEY = app.secret_key
 else:
@@ -94,28 +98,7 @@ TOTP = True
 IGNORE_SSL = False
 SSL_DEFAULT = ssl._create_default_https_context
 SSL_UNVERIFIED = ssl._create_unverified_context
-
-class ReverseProxied(object):
-    def __init__(self, app, script_name=None, scheme=None, server=None):
-        self.app = app
-        self.script_name = script_name
-        self.scheme = scheme
-        self.server = server
-
-    def __call__(self, environ, start_response):
-        script_name = environ.get('HTTP_X_SCRIPT_NAME', '') or self.script_name
-        if script_name:
-            environ['SCRIPT_NAME'] = script_name
-            path_info = environ['PATH_INFO']
-            if path_info.startswith(script_name):
-                environ['PATH_INFO'] = path_info[len(script_name):]
-        scheme = environ.get('HTTP_X_SCHEME', '') or self.scheme
-        if scheme:
-            environ['wsgi.url_scheme'] = scheme
-        server = environ.get('HTTP_X_FORWARDED_SERVER', '') or self.server
-        if server:
-            environ['HTTP_HOST'] = server
-        return self.app(environ, start_response)
+IS_GUNICORN = "gunicorn" in os.environ.get("SERVER_SOFTWARE", "")
 
 def load_settings():
     global TITLE, SESSION_TIMEOUT, BANLIMIT, BANTIME, MAXFIDO, TOTP, IGNORE_SSL, ssl
@@ -145,13 +128,13 @@ def checkban(addr):
             banned.delete()
         elif banned.failed_attempts > BANLIMIT and time.time() - banned.last_attempt < BANTIME:
             banned.increment()
-            print("Denying access from:", addr)
+            app.logger.warning("Denying access from: %s", addr)
             return False
     return True
 
 
 def add_to_ban(addr):
-    print("Adding to banlist:", addr)
+    app.logger.warning("Adding to banlist: %s", addr)
     banned = Banlist.query.filter_by(ip=request.remote_addr).first()
     if not banned:
         banned = Banlist(
@@ -168,7 +151,7 @@ def add_to_ban(addr):
 def unban(addr):
     banned = Banlist.query.filter_by(ip=request.remote_addr).first()
     if banned:
-        print("Removing from banlist:", addr)
+        app.logger.warning("Removing from banlist: %s", addr)
         banned.delete()
 
 @login_manager.user_loader
@@ -182,6 +165,10 @@ def load_user(user_id):
 
 @app.before_request
 def before_request():
+    if ORIGIN not in request.host_url:
+        app.logger.warning('Incorrect hostname (%s), rejecting request', request.host_url)
+        add_to_ban(request.remote_addr)
+        abort(401)
     session.permanent = True
     app.permanent_session_lifetime = datetime.timedelta(minutes=SESSION_TIMEOUT)
     session.modified = True
@@ -189,7 +176,7 @@ def before_request():
 
 @app.context_processor
 def inject_dict_for_all_templates():
-    return dict(app_title=TITLE, debug=app.debug)
+    return dict(app_title=TITLE, debug=app.debug, version=VERSION)
 
 
 @app.template_filter('ctime')
@@ -197,6 +184,11 @@ def timectime(s):
     return time.ctime(s)
 
 
+@app.route('/manifest.json')
+def pwa_manifest():
+    return render_template('pwa-manifest.json')
+
+
 @app.route('/index')
 @app.route('/')
 def index():
@@ -207,7 +199,7 @@ def index():
     users = User.query.all()
     if not users:
         return redirect(url_for('register_prompt', reg_token="none"))
-    return render_template('index.html')
+    return render_template('index.html', totp=TOTP)
 
 
 @app.route('/about')
@@ -243,6 +235,9 @@ def settings():
         ignore_ssl.value = '1' if request.values.get('ignore_ssl') else '0'
         db.session.add(ignore_ssl)
         db.session.commit()
+        if IS_GUNICORN:
+            app.logger.warning("Sending HUP to %i", os.getppid())
+            os.kill(os.getppid(), signal.SIGHUP)
         load_settings()
     return render_template('settings.html', title=TITLE,
                            session_timeout=SESSION_TIMEOUT,
@@ -262,6 +257,11 @@ def banlist():
         if banned:
             banned.delete()
         return redirect(url_for('banlist'))
+    purge = request.args.get('purge')
+    if purge:
+        for ip in Banlist.query.all():
+            ip.delete()
+        return redirect(url_for('banlist'))
     banlist = Banlist.query.all()
     return render_template('banlist.html', banlist=banlist)
 
@@ -274,6 +274,7 @@ def register_prompt(reg_token):
         token = RegToken.query.filter_by(token=reg_token).first()
         if not token:
             add_to_ban(request.remote_addr)
+            abort(401)
     if current_user.is_authenticated:
         return redirect(url_for('triggers'))
     return render_template('register.html', reg_token=reg_token)
@@ -293,7 +294,7 @@ def register():
     if User.query.all():
         token = RegToken.query.filter_by(token=reg_token).first()
         if not token:
-            print("Invalid token")
+            app.logger.warning("Invalid token")
             add_to_ban(request.remote_addr)
             return make_response(jsonify({'status': 'error'}), 401)
         if token:
@@ -329,7 +330,54 @@ def register():
     if token:
         token.delete()
     login_user(user)
-    return redirect(url_for('zfa'))
+    return redirect(url_for('security'))
+
+
+@app.route('/users/import', methods=['POST'])
+@login_required
+def users_import():
+    if not checkban(request.remote_addr):
+        abort(401)
+    if not current_user.is_admin:
+        redirect(url_for('triggers'))
+    data = request.get_json()
+    user = User(
+        created=data['created'],
+        username=data['username'],
+        display_name=data['display_name'],
+        is_admin=data['is_admin'],
+        otp_only=data['otp_only'],
+        password_hash=data['password_hash'],
+        sign_count=data['sign_count'],
+        last_failed=data['last_failed'],
+        last_login=data['last_login'],
+        totp_enabled=data['totp_enabled'],
+        totp_initialized=data['totp_initialized'],
+        totp_secret=data['totp_secret'].encode('utf-8'),
+        icon_url=data['icon_url'])
+    db.session.add(user)
+    db.session.commit()
+
+    return make_response(jsonify({"status": "success"}), 200)
+
+
+@app.route('/authenticators/import', methods=['POST'])
+@login_required
+def authenticators_import():
+    if not checkban(request.remote_addr):
+        abort(401)
+    if not current_user.is_admin:
+        redirect(url_for('triggers'))
+    data = request.get_json()
+    app.logger.debug(data)
+    authenticator = Authenticator(
+        credential=base64.b64decode(data['credential'].encode('utf-8')),
+        user=int(data['user']),
+        name=data['name'])
+    db.session.add(authenticator)
+    db.session.commit()
+
+    return make_response(jsonify({"status": "success"}), 200)
 
 
 ### This is only accessible when Flasks Debug mode is turned on!
@@ -357,7 +405,7 @@ def login_debug():
     db.session.add(user)
     db.session.commit()
     unban(request.remote_addr)
-    return redirect(url_for('zfa'))
+    return redirect(url_for('security'))
 
 
 @app.route("/login/otp", methods=["POST"])
@@ -371,19 +419,19 @@ def login_otp():
 
     if not util.validate_username(username):
         add_to_ban(request.remote_addr)
-        print("Invalid username")
+        app.logger.warning("Invalid username")
         return make_response(jsonify({'status': 'error'}), 401)
 
     user = User.query.filter_by(username=username).first()
 
     if not user:
         add_to_ban(request.remote_addr)
-        print("User does not exist")
+        app.logger.warning("User does not exist")
         return make_response(jsonify({'status': 'error'}), 401)
     if not user.check_password(password):
         user.failed()
         add_to_ban(request.remote_addr)
-        print("Wrong password")
+        app.logger.warning("Wrong password")
         return make_response(jsonify({'status': 'error'}), 401)
 
     otps = []
@@ -393,7 +441,7 @@ def login_otp():
     if not user.totp_secret and not otps:
         user.failed()
         add_to_ban(request.remote_addr)
-        print("No (T)OTP available")
+        app.logger.warning("No (T)OTP available")
         return make_response(jsonify({'status': 'error'}), 401)
 
     if TOTP and totp:
@@ -414,8 +462,8 @@ def login_otp():
                     unban(request.remote_addr)
                     return make_response(jsonify({'status': 'success'}), 200)
             except InvalidToken:
-                print("Invalid TOTP. Server secret key may has changed")
-            print("Incorrect TOTP")
+                app.logger.warning("Invalid TOTP. Server secret key may has changed")
+            app.logger.warning("Incorrect TOTP")
 
     if otp and otps:
         now = int(time.time())
@@ -429,16 +477,27 @@ def login_otp():
                 db.session.commit()
                 unban(request.remote_addr)
                 return make_response(jsonify({'status': 'success'}), 200)
-        print("Invalid OTP")
+        app.logger.warning("Invalid OTP")
 
     user.failed()
     add_to_ban(request.remote_addr)
     return make_response(jsonify({'status': 'error'}), 401)
 
 
-@app.route('/zfa', methods=['GET'])
+@app.route('/security', methods=['GET', 'POST'])
 @login_required
-def zfa():
+def security():
+    user = User.query.filter_by(username=current_user.username).first()
+    if request.method == 'POST':
+        current_password = request.form.get('current_password')
+        if not user.check_password(current_password):
+            flash("Incorrect password.")
+            return redirect(url_for('security'))
+        user.password_hash = generate_password_hash(request.form.get('password'))
+        db.session.add(user)
+        db.session.commit()
+        flash("Password saved.")
+        return redirect(url_for('security'))
     totp_enabled = bool(int(Setting.query.filter_by(parameter='totp').first().value))
     del_authenticator = request.args.get('del_authenticator')
     if del_authenticator:
@@ -448,11 +507,10 @@ def zfa():
             authenticator = Authenticator.query.filter_by(id=del_authenticator).filter_by(user=current_user.id).first()
         if authenticator:
             authenticator.delete()
-        return redirect(url_for('zfa'))
+        return redirect(url_for('security'))
     authenticators = []
     for authenticator in Authenticator.query.filter_by(user=current_user.id):
         authenticators.append({'id': authenticator.id, 'name': authenticator.name, 'user': authenticator.user})
-    user = User.query.filter_by(username=current_user.username).first()
     kdf = PBKDF2HMAC(algorithm=hashes.SHA256(), length=32,
                      salt=user.username.encode("utf-8"),
                      iterations=100000, backend=default_backend())
@@ -466,14 +524,14 @@ def zfa():
             totp_uri = pyotp.totp.TOTP(totp_secret).provisioning_uri(
                 name=current_user.username, issuer_name=TITLE)
         except InvalidToken:
-            print("Invalid token. Server secret key may has changed.")
+            app.logger.warning("Invalid token. Server secret key may has changed.")
             totp_secret = "Invalid"
     if not totp_secret == "Invalid":
         if totp_secret is not None and user.totp_initialized:
             totp_secret = "Initialized"
             totp_uri = ""
 
-    return render_template('2fa.html', authenticators=authenticators,
+    return render_template('security.html', authenticators=authenticators,
                            totp_secret=totp_secret, totp_uri=totp_uri,
                            totp_enabled=totp_enabled, user=user, maxfido=MAXFIDO)
 
@@ -592,6 +650,17 @@ def totp_delete_username(username):
     return make_response(jsonify({"status": "success"}), 200)
 
 
+@app.route('/authenticators/<int:authenticatorid>')
+@login_required
+def authenticators_json(authenticatorid):
+    authenticator = Authenticator.query.filter_by(id=authenticatorid).first()
+    authenticator = {
+        "name": authenticator.name,
+        "credential": base64.b64encode(authenticator.credential).decode('utf-8')
+    }
+    return make_response(jsonify(authenticator), 200)
+
+
 @app.route('/users', methods=['GET'])
 @login_required
 def users():
@@ -604,6 +673,8 @@ def users():
             for authenticator in Authenticator.query.filter_by(user=user.id):
                 authenticator.delete()
             user.delete()
+            TriggerUserMap.query.filter_by(user=user.id).delete()
+            db.session.commit()
         return redirect(url_for('users'))
     users = User.query.all()
     usermap = {}
@@ -615,6 +686,30 @@ def users():
     return render_template('users.html', users=users, usermap=usermap, authenticators=authenticators)
 
 
+@app.route('/users/<int:userid>')
+@login_required
+def users_json(userid):
+    user = User.query.filter_by(id=userid).first()
+    userdata = {
+        "username": user.username,
+        "display_name": user.display_name,
+        "created": user.created,
+        "password_hash": user.password_hash,
+        "is_admin": user.is_admin,
+        "sign_count": user.sign_count,
+        "failed_logins": user.failed_logins,
+        "last_login": user.last_login,
+        "last_failed": user.last_failed,
+        "icon_url": user.icon_url,
+        "totp_enabled": user.totp_enabled,
+        "totp_initialized": user.totp_initialized,
+        "otp_only": user.otp_only
+    }
+    if user.totp_secret is not None:
+        userdata["totp_secret"] = user.totp_secret.decode('utf-8')
+    return make_response(jsonify(userdata), 200)
+
+
 @app.route('/users/toggle_admin/<int:userid>')
 @login_required
 def users_toggle_admin(userid):
@@ -654,7 +749,9 @@ def users_toggle_totp(userid):
 @app.route('/triggers')
 @login_required
 def triggers():
-    triggers = Trigger.query.order_by(Trigger.order.asc()).all()
+    available_triggers = [ x.trigger for x in TriggerUserMap.query.filter_by(user=current_user.id).all() ]
+    triggers = Trigger.query.filter_by(disabled=False).order_by(Trigger.order.asc()).all()
+    triggers = [ t for t in triggers if t.id in available_triggers ]
     return render_template('triggers.html', triggers=triggers)
 
 
@@ -668,9 +765,15 @@ def triggers_json(triggerid):
         "order": trigger.order,
         "trigger_json": trigger.trigger_json,
         "include_user": trigger.include_user,
+        "require_geo": trigger.require_geo,
+        "disabled": trigger.disabled,
         "webhook_uri": trigger.webhook_uri,
         "password": trigger.password,
+        "users": []
     }
+    usermap = TriggerUserMap.query.filter_by(trigger=triggerid).all()
+    for user in usermap:
+        trigger["users"].append(user.user)
     return make_response(jsonify(trigger), 200)
 
 
@@ -678,8 +781,16 @@ def triggers_json(triggerid):
 @login_required
 def triggers_fire(triggerid):
     trigger = Trigger.query.filter_by(id=triggerid).first()
-    password = request.values.get('password')
-    if password:
+    if trigger.disabled:
+        return make_response(jsonify({"status": "error", "error": "trigger disabled"}), 401)
+    available_triggers = [ x.trigger for x in TriggerUserMap.query.filter_by(user=current_user.id).all() ]
+    if trigger.id not in available_triggers:
+        return make_response(jsonify({"status": "error", "error": "trigger not available to user"}), 401)
+    postdata = request.get_json()
+    password = postdata.get('password')
+    if trigger.password and not password:
+        return make_response(jsonify({"status": "error", "error": "missing password"}), 401)
+    if password and trigger.password:
         if password != trigger.password:
             return make_response(jsonify({"status": "error", "error": "invalid password"}), 401)
     headers = {
@@ -688,17 +799,21 @@ def triggers_fire(triggerid):
     data = json.loads(trigger.trigger_json)
     if trigger.include_user:
         data['user'] = current_user.username
-    print("Trigger fired:", trigger.caption)
+    if trigger.require_geo:
+        data['latitude'] = postdata.get('latitude')
+        data['longitude'] = postdata.get('longitude')
+        data['accuracy'] = postdata.get('accuracy')
+    app.logger.warning("Trigger fired by %s: %s", current_user.username, trigger.caption)
     req = urllib.request.Request(trigger.webhook_uri,
                                  headers=headers, method='POST',
                                  data=bytes(json.dumps(data).encode('utf-8')))
     try:
         with urllib.request.urlopen(req) as response:
             if response.code != 200:
-                print("Trigger failed:" % trigger.caption)
+                app.logger.warning("Trigger failed: %s", trigger.caption)
                 return make_response(jsonify({"status": "failed", "trigger": trigger.id}), 200)
     except Exception as err:
-        print(err)
+        app.logger.warning(err)
         return make_response(jsonify({"status": "failed", "trigger": trigger.id}), 200)
     return make_response(jsonify({"status": "success", "trigger": trigger.id}), 200)
 
@@ -712,6 +827,8 @@ def admin_triggers():
             trigger = Trigger.query.filter_by(id=del_trigger).first()
             if trigger:
                 trigger.delete()
+                TriggerUserMap.query.filter_by(trigger=del_trigger).delete()
+                db.session.commit()
             return redirect(url_for('admin_triggers'))
     elif request.method == 'POST':
         trigger_id = request.values.get('id')
@@ -719,8 +836,11 @@ def admin_triggers():
         order = int(request.values.get('order'))
         trigger_json = json.loads(request.values.get('trigger_json'))
         include_user = True if request.values.get('include_user') else False
+        require_geo = True if request.values.get('require_geo') else False
+        disabled = True if request.values.get('disabled') else False
         webhook_uri = request.values.get('webhook_uri')
         password = request.values.get('password')
+        users = request.form.getlist('users')
         trigger_json = json.dumps(trigger_json)
         if not trigger_id:
             trigger = Trigger(
@@ -728,35 +848,41 @@ def admin_triggers():
                 order=order,
                 trigger_json=trigger_json,
                 include_user=include_user,
+                require_geo=require_geo,
+                disabled=disabled,
                 webhook_uri=webhook_uri,
                 password=password
             )
             db.session.add(trigger)
             db.session.commit()
-        else:
-            if Trigger.query.get(int(trigger_id)):
-                trigger = Trigger.query.get(int(trigger_id))
-                trigger.caption = caption
-                trigger.order = order
-                trigger.trigger_json = trigger_json
-                trigger.include_user = include_user
-                trigger.webhook_uri = webhook_uri
-                trigger.password = password
-                db.session.add(trigger)
-                db.session.commit()
-            else:
-                trigger = Trigger(
-                    caption=caption,
-                    order=order,
-                    trigger_json=trigger_json,
-                    include_user=include_user,
-                    webhook_uri=webhook_uri,
-                    password=password
+            for user in users:
+                trigger_user = TriggerUserMap(
+                    trigger=int(trigger.id),
+                    user=int(user)
                 )
-                db.session.add(trigger)
-                db.session.commit()
+                db.session.add(trigger_user)
+        else:
+            trigger = Trigger.query.get(int(trigger_id))
+            trigger.caption = caption
+            trigger.order = order
+            trigger.trigger_json = trigger_json
+            trigger.include_user = include_user
+            trigger.require_geo = require_geo
+            trigger.disabled = disabled
+            trigger.webhook_uri = webhook_uri
+            trigger.password = password
+            db.session.add(trigger)
+            TriggerUserMap.query.filter_by(trigger=trigger.id).delete()
+            for user in users:
+                trigger_user = TriggerUserMap(
+                    trigger=int(trigger.id),
+                    user=int(user)
+                )
+                db.session.add(trigger_user)
+        db.session.commit()
     triggers = Trigger.query.all()
-    return render_template('admin_triggers.html', triggers=triggers)
+    users = User.query.all()
+    return render_template('admin_triggers.html', triggers=triggers, users=users)
 
 
 @app.route("/api/register/begin", methods=["POST"])
@@ -765,7 +891,7 @@ def register_begin():
     user = load_user(current_user.id)
     authenticators = Authenticator.query.filter_by(user=user.id).all()
     if len(authenticators) >= MAXFIDO:
-        print("No additional tokens allowed")
+        app.logger.warning("No additional tokens allowed")
         return make_response(jsonify({'fail': 'error'}), 401)
     authenticators = []
     registration_data, state = server.register_begin(
@@ -817,18 +943,18 @@ def authenticate_begin():
     password = request.form.get('login_password')
 
     if not util.validate_username(username):
-        print("Invalid username")
+        app.logger.warning("Invalid username")
         add_to_ban(request.remote_addr)
         return make_response(jsonify({'fail': 'error'}), 401)
 
     user = User.query.filter_by(username=username).first()
 
     if not user:
-        print("User does not exist")
+        app.logger.warning("User does not exist")
         add_to_ban(request.remote_addr)
         return make_response(jsonify({'fail': 'error'}), 401)
     if not user.check_password(password):
-        print("Incorrect password")
+        app.logger.warning("Incorrect password")
         user.failed()
         add_to_ban(request.remote_addr)
         return make_response(jsonify({'fail': 'error'}), 401)
@@ -837,7 +963,7 @@ def authenticate_begin():
     for authenticator in Authenticator.query.filter_by(user=user.id):
         authenticators.append(AttestedCredentialData(authenticator.credential))
     if not authenticators:
-        print("No authenticator enrolled")
+        app.logger.warning("No authenticator enrolled")
         return make_response(jsonify({'fail': 'error'}), 401)
     auth_data, state = server.authenticate_begin(authenticators)
     session["state"] = state
